package com.uas.platform.core.data.repository;

import java.io.Serializable;
import java.util.List;

import org.springframework.data.repository.NoRepositoryBean;
import org.springframework.data.repository.Repository;

/**
 * 调用了继承该接口的Repository的方法会在操作完成之后自动添加日志 <br>
 * 日志所需参数在实体用注解标志
 * 
 * @see Repository
 * @see OperationLogger
 * @see OperationLog
 * 
 * @author yingp
 * 
 * @param <T>
 *            Entity
 * @param <ID>
 *            ID
 */
@NoRepositoryBean
public abstract interface LogRepository<T, ID extends Serializable> extends Repository<T, ID> {

	/**
	 * 保存、修改
	 * 
	 * @param paramS
	 * @return
	 */
	public abstract <S extends T> S saveAndLog(S paramS);

	/**
	 * 批量保存、修改
	 * 
	 * @param paramS
	 * @return
	 */
	public abstract <S extends T> List<S> saveAndLog(Iterable<S> paramS);

	/**
	 * 修改
	 * 
	 * @param paramS
	 * @return
	 */
	public abstract <S extends T> S updateAndLog(S paramS);

	/**
	 * 批量修改
	 * 
	 * @param paramS
	 * @return
	 */
	public abstract <S extends T> List<S> updateAndLog(Iterable<S> paramS);

	/**
	 * 提交
	 * 
	 * @param paramS
	 * @return
	 */
	public abstract <S extends T> S submitAndLog(S paramS);

	/**
	 * 提交
	 * 
	 * @param id
	 *            实体id
	 * @return
	 */
	public abstract T submitAndLog(ID id);

	/**
	 * 批量提交
	 * 
	 * @param paramS
	 *            实体
	 * @return
	 */
	public abstract <S extends T> List<S> submitAndLog(Iterable<S> paramS);

	/**
	 * 审批
	 * 
	 * @param paramS
	 * @return
	 */
	public abstract <S extends T> S auditAndLog(S paramS);

	/**
	 * 审批
	 * 
	 * @param id
	 *            实体id
	 * @return
	 */
	public abstract T auditAndLog(ID id);

	/**
	 * 批量审批
	 * 
	 * @param paramS
	 *            实体
	 * @return
	 */
	public abstract <S extends T> List<S> auditAndLog(Iterable<S> paramS);
	
	/**
	 * 反审批
	 * 
	 * @param paramS
	 * @return
	 */
	public abstract <S extends T> S unAuditAndLog(S paramS);

	/**
	 * 反审批
	 * 
	 * @param id
	 *            实体id
	 * @return
	 */
	public abstract T unAuditAndLog(ID id);

	/**
	 * 批量反审批
	 * 
	 * @param paramS
	 *            实体
	 * @return
	 */
	public abstract <S extends T> List<S> unAuditAndLog(Iterable<S> paramS);

	/**
	 * 回复
	 * 
	 * @param paramS
	 *            实体
	 * @return
	 */
	public abstract <S extends T> S replyAndLog(S paramS);

	/**
	 * 回复
	 * 
	 * @param id
	 *            实体ID
	 * @return
	 */
	public abstract T replyAndLog(ID id);

	/**
	 * 批量回复
	 * 
	 * @param paramS
	 *            实体
	 * @return
	 */
	public abstract <S extends T> List<S> replyAndLog(Iterable<S> paramS);

	/**
	 * 删除
	 * 
	 * @param id
	 *            实体ID
	 */
	public abstract void deleteAndLog(ID id);

	/**
	 * 删除
	 * 
	 * @param paramT
	 */
	public abstract void deleteAndLog(T paramT);

	/**
	 * 批量删除
	 * 
	 * @param paramS
	 */
	public abstract void deleteAndLog(Iterable<? extends T> paramS);

}
