package com.uas.platform.core.data.repository;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.List;

import javax.persistence.EntityManager;
import javax.persistence.NoResultException;
import javax.persistence.TypedQuery;
import javax.persistence.criteria.*;

import org.springframework.data.domain.Page;
import org.springframework.data.domain.Pageable;
import org.springframework.data.domain.Sort;
import org.springframework.data.jpa.domain.Specification;
import org.springframework.data.jpa.repository.EntityGraph.EntityGraphType;
import org.springframework.data.jpa.repository.support.JpaEntityInformation;
import org.springframework.data.jpa.repository.support.JpaEntityInformationSupport;
import org.springframework.data.jpa.repository.support.SimpleJpaRepository;
import org.springframework.data.repository.NoRepositoryBean;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.util.Assert;

import com.uas.platform.core.data.support.OperationLogger;
import com.uas.platform.core.model.Operation;

/**
 * 
 * @see GenericRepository
 * @see LogRepository
 * @see SimpleJpaRepository
 * 
 * @author yingp
 * 
 * @param <T>
 * @param <ID>
 */
@NoRepositoryBean
public class SimpleGenericRepository<T, ID extends Serializable> extends SimpleJpaRepository<T, ID> implements
		GenericRepository<T, ID> {

	public SimpleGenericRepository(Class<T> domainClass, EntityManager em) {
		super(domainClass, em);
		this.em = em;
		this.logger = OperationLogger.getLogger(domainClass);
		this.entityInformation = JpaEntityInformationSupport.getEntityInformation(domainClass, em);
	}

	private final EntityManager em;
	private final JpaEntityInformation<T, ?> entityInformation;
	private final OperationLogger logger;

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * com.uas.scm.common.data.repository.LogRepository#saveAndLog(java.lang
	 * .Object)
	 */
	@Transactional
	public <S extends T> S saveAndLog(S entity) {
		if (this.entityInformation.isNew(entity)) {
			this.logger.status(entity, Operation.INSERT);
			em.persist(entity);
			em.persist(logger.insert(entity));
			return entity;
		}
		return updateAndLog(entity);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * com.uas.scm.common.data.repository.LogRepository#saveAndLog(java.lang
	 * .Iterable)
	 */
	@Transactional
	public <S extends T> List<S> saveAndLog(Iterable<S> entities) {
		List<S> result = new ArrayList<S>();
		if (entities != null) {
			for (S entity : entities) {
				result.add(saveAndLog(entity));
			}
			flush();
		}
		return result;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * com.uas.scm.common.data.repository.LogRepository#updateAndLog(java.lang
	 * .Object)
	 */
	@Transactional
	public <S extends T> S updateAndLog(S entity) {
		Assert.notNull(entity);
		entity = em.merge(entity);
		em.persist(logger.update(entity));
		return entity;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * com.uas.scm.common.data.repository.LogRepository#updateAndLog(java.lang
	 * .Iterable)
	 */
	@Transactional
	public <S extends T> List<S> updateAndLog(Iterable<S> entities) {
		List<S> result = new ArrayList<S>();
		if (entities != null) {
			for (S entity : entities) {
				result.add(updateAndLog(entity));
			}
		}
		return result;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * com.uas.scm.common.data.repository.LogRepository#submitAndLog(java.lang
	 * .Object)
	 */
	@Transactional
	public <S extends T> S submitAndLog(S entity) {
		Assert.notNull(entity);
		em.persist(logger.submit(entity));
		return em.merge(entity);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * com.uas.scm.common.data.repository.LogRepository#submitAndLog(java.io
	 * .Serializable)
	 */
	@Transactional
	public T submitAndLog(ID id) {
		Assert.notNull(id, "The given id must not be null!");
		T entity = findOne(id);
		if (entity != null) {
			return submitAndLog(entity);
		}
		return null;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * com.uas.scm.common.data.repository.LogRepository#submitAndLog(java.lang
	 * .Iterable)
	 */
	@Transactional
	public <S extends T> List<S> submitAndLog(Iterable<S> entities) {
		List<S> result = new ArrayList<S>();
		if (entities != null) {
			for (S entity : entities) {
				result.add(submitAndLog(entity));
			}
		}
		return result;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * com.uas.scm.common.data.repository.LogRepository#auditAndLog(java.lang
	 * .Object)
	 */
	@Transactional
	public <S extends T> S auditAndLog(S entity) {
		Assert.notNull(entity);
		em.persist(logger.audit(entity));
		return em.merge(entity);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * com.uas.scm.common.data.repository.LogRepository#auditAndLog(java.io.
	 * Serializable)
	 */
	@Transactional
	public T auditAndLog(ID id) {
		Assert.notNull(id, "The given id must not be null!");
		T entity = findOne(id);
		if (entity != null) {
			return auditAndLog(entity);
		}
		return null;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * com.uas.scm.common.data.repository.LogRepository#auditAndLog(java.lang
	 * .Iterable)
	 */
	@Transactional
	public <S extends T> List<S> auditAndLog(Iterable<S> entities) {
		List<S> result = new ArrayList<S>();
		if (entities != null) {
			for (S entity : entities) {
				result.add(auditAndLog(entity));
			}
		}
		return result;
	}

	@Transactional
	public <S extends T> S unAuditAndLog(S entity) {
		Assert.notNull(entity);
		em.persist(logger.unAudit(entity));
		return em.merge(entity);
	}

	@Transactional
	public T unAuditAndLog(ID id) {
		Assert.notNull(id, "The given id must not be null!");
		T entity = findOne(id);
		if (entity != null) {
			return unAuditAndLog(entity);
		}
		return null;
	}

	@Transactional
	public <S extends T> List<S> unAuditAndLog(Iterable<S> entities) {
		List<S> result = new ArrayList<S>();
		if (entities != null) {
			for (S entity : entities) {
				result.add(unAuditAndLog(entity));
			}
		}
		return result;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * com.uas.scm.common.data.repository.LogRepository#replyAndLog(java.lang
	 * .Object)
	 */
	@Transactional
	public <S extends T> S replyAndLog(S entity) {
		Assert.notNull(entity);
		em.persist(logger.reply(entity));
		return em.merge(entity);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * com.uas.scm.common.data.repository.LogRepository#replyAndLog(java.io.
	 * Serializable)
	 */
	@Transactional
	public T replyAndLog(ID id) {
		Assert.notNull(id, "The given id must not be null!");
		T entity = findOne(id);
		if (entity != null) {
			return replyAndLog(entity);
		}
		return null;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * com.uas.scm.common.data.repository.LogRepository#replyAndLog(java.lang
	 * .Iterable)
	 */
	@Transactional
	public <S extends T> List<S> replyAndLog(Iterable<S> entities) {
		List<S> result = new ArrayList<S>();
		if (entities != null) {
			for (S entity : entities) {
				result.add(replyAndLog(entity));
			}
		}
		return result;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * com.uas.scm.common.data.repository.LogRepository#deleteAndLog(java.lang
	 * .Object)
	 */
	@Transactional
	public void deleteAndLog(T entity) {
		Assert.notNull(entity, "The entity must not be null!");
		em.persist(logger.delete(entity));
		em.remove(em.contains(entity) ? entity : em.merge(entity));
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * com.uas.scm.common.data.repository.LogRepository#deleteAndLog(java.lang
	 * .Iterable)
	 */
	@Transactional
	public void deleteAndLog(Iterable<? extends T> entities) {
		if (entities != null) {
			for (T entity : entities) {
				deleteAndLog(entity);
			}
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * com.uas.scm.common.data.repository.LogRepository#deleteAndLog(java.io
	 * .Serializable)
	 */
	@Transactional
	public void deleteAndLog(ID id) {
		Assert.notNull(id, "The given id must not be null!");
		T entity = findOne(id);
		if (entity != null)
			deleteAndLog(entity);
	}

	@Override
	public List<T> findAll(Specification<T> spec, EntityGraphType entityGraphType, String entityGraphName) {
		TypedQuery<T> query = getQuery(spec, (Sort) null);
		query.setHint(entityGraphType.getKey(), em.getEntityGraph(entityGraphName));
		return query.getResultList();
	}

	@Override
	public Page<T> findAll(Specification<T> spec, Pageable pageable, EntityGraphType entityGraphType, String entityGraphName) {
		TypedQuery<T> query = getQuery(spec, pageable.getSort());
		query.setHint(entityGraphType.getKey(), em.getEntityGraph(entityGraphName));
		return readPage(query, pageable, spec);
	}

	@Override
	public List<T> findAll(Specification<T> spec, Sort sort, EntityGraphType entityGraphType, String entityGraphName) {
		TypedQuery<T> query = getQuery(spec, sort);
		query.setHint(entityGraphType.getKey(), em.getEntityGraph(entityGraphName));
		return query.getResultList();
	}

	@Override
	public List<T> findAll(Iterable<ID> ids, EntityGraphType entityGraphType, String entityGraphName) {
		if (ids == null || !ids.iterator().hasNext()) {
			return Collections.emptyList();
		}

		if (entityInformation.hasCompositeId()) {

			List<T> results = new ArrayList<T>();

			for (ID id : ids) {
				results.add(findOne(id));
			}

			return results;
		}

		ByIdsSpecification<T> specification = new ByIdsSpecification<T>(entityInformation);
		TypedQuery<T> query = getQuery(specification, (Sort) null);
		query.setParameter(specification.parameter, ids);
		query.setHint(entityGraphType.getKey(), em.getEntityGraph(entityGraphName));
		return query.getResultList();
	}

	@Override
	public List<T> findAll(Collection<ID> ids, EntityGraphType entityGraphType, String entityGraphName) {
		IdsSpecification<T, ID> specification = new IdsSpecification<>(entityInformation, ids);
		TypedQuery<T> query = getQuery(specification, (Sort) null);
		query.setHint(entityGraphType.getKey(), em.getEntityGraph(entityGraphName));
		return query.getResultList();
	}

	@Override
	public T findOne(Specification<T> spec, EntityGraphType entityGraphType, String entityGraphName) {
		try {
			TypedQuery<T> query = getQuery(spec, (Sort) null);
			query.setHint(entityGraphType.getKey(), em.getEntityGraph(entityGraphName));
			return query.getSingleResult();
		} catch (NoResultException e) {
			return null;
		}
	}

	private static final class ByIdsSpecification<T> implements Specification<T> {

		private final JpaEntityInformation<T, ?> entityInformation;

		ParameterExpression<Iterable> parameter;

		public ByIdsSpecification(JpaEntityInformation<T, ?> entityInformation) {
			this.entityInformation = entityInformation;
		}

		public Predicate toPredicate(Root<T> root, CriteriaQuery<?> query, CriteriaBuilder cb) {
			Path<?> path = root.get(entityInformation.getIdAttribute());
			parameter = cb.parameter(Iterable.class);
			return path.in(parameter);
		}
	}

	private static final class IdsSpecification<T, ID> implements Specification<T> {

		private final JpaEntityInformation<T, ?> entityInformation;

		private final Collection<ID> ids;

		public IdsSpecification(JpaEntityInformation<T, ?> entityInformation, Collection<ID> ids) {
			this.entityInformation = entityInformation;
			this.ids = ids;
		}

		public Predicate toPredicate(Root<T> root, CriteriaQuery<?> query, CriteriaBuilder cb) {
			Path<?> path = root.get(entityInformation.getIdAttribute());
			return path.in(ids);
		}
	}
}
